@@ -101,9 +101,10 @@
     (:translate symbol-value)
     (:args (symbol :scs (descriptor-reg) :to :save
                    :load-if (not (and (sc-is symbol constant)
-                                      (or (symbol-always-has-tls-value-p (tn-value symbol))
-                                          (symbol-always-has-tls-index-p (tn-value symbol)))))))
-    (:args-var symbol-tn-ref)
+                                      (let ((pkg (sb-xc:symbol-package (tn-value symbol))))
+                                        (or (and pkg (system-package-p pkg))
+                                            (eq pkg *cl-package*)))))))
+    (:arg-refs symbol-tn-ref)
     (:temporary (:sc any-reg) tls-index)
     (:variant-vars check-boundp)
     (:variant t)
@@ -210,7 +211,7 @@
   (:policy :fast-safe)
   (:translate symbol-hash)
   (:args (symbol :scs (descriptor-reg)))
-  (:args-var args)
+  (:arg-refs args)
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
   (:generator 2
@@ -226,10 +227,11 @@
   (:args (symbol :scs (descriptor-reg)))
   (:results (result :scs (unsigned-reg)))
   (:result-types positive-fixnum)
-  (:translate sb-impl::symbol-package-id)
+  (:translate symbol-package-id)
   (:policy :fast-safe)
-  (:generator 1 ; ASSUMPTION: symbol-package-bits = 16
-   (inst ldrh result (@ symbol (+ (ash symbol-name-slot word-shift)
+  (:generator 1
+    #.(assert (= sb-impl::package-id-bits 16))
+    (inst ldrh result (@ symbol (+ (ash symbol-name-slot word-shift)
                                   (- other-pointer-lowtag)
                                   6))))) ; little-endian
 (define-vop ()
@@ -238,8 +240,9 @@
   (:args (symbol :scs (descriptor-reg)))
   (:results (result :scs (descriptor-reg)))
   (:generator 5
-    (loadw tmp-tn symbol symbol-name-slot other-pointer-lowtag)
-    (inst and result tmp-tn (1- (ash 1 sb-impl::symbol-name-bits)))))
+    (pseudo-atomic (tmp-tn :sync nil)
+      (loadw tmp-tn symbol symbol-name-slot other-pointer-lowtag)
+      (inst and result tmp-tn (1- (ash 1 sb-impl::symbol-name-bits))))))
 
 (define-vop (%compare-and-swap-symbol-value)
   (:translate %compare-and-swap-symbol-value)
@@ -378,9 +381,6 @@
   (define-vop (dynbind)
     (:args (value :scs (any-reg descriptor-reg zero) :to :save)
            (symbol :scs (descriptor-reg)
-                   :load-if (not (and (sc-is symbol constant)
-                                      (or (symbol-always-has-tls-value-p (tn-value symbol))
-                                          (symbol-always-has-tls-index-p (tn-value symbol)))))
                    :target alloc-tls-symbol))
     (:temporary (:sc descriptor-reg :offset r8-offset :from (:argument 1)) alloc-tls-symbol)
     (:temporary (:sc non-descriptor-reg :offset nl0-offset) tls-index)
@@ -388,28 +388,59 @@
     (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
     (:ignore free-tls-index)
     (:temporary (:scs (any-reg)) bsp)
-     (:generator 5
+    (:generator 5
       (load-binding-stack-pointer bsp)
       (inst add bsp bsp (* binding-size n-word-bytes))
       (store-binding-stack-pointer bsp)
-      (let* ((known-symbol-p (sc-is symbol constant))
-             (known-symbol (and known-symbol-p (tn-value symbol)))
+      (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
+      (inst cbnz (32-bit-reg tls-index) TLS-INDEX-VALID)
+      (move alloc-tls-symbol symbol)
+      (load-inline-constant lr '(:fixup alloc-tls-index :assembly-routine))
+      (inst blr lr)
+      TLS-INDEX-VALID
+      (inst ldr alloc-tls-symbol (@ thread-tn tls-index))
+      (inst stp alloc-tls-symbol tls-index
+        (@ bsp (* (- binding-value-slot binding-size)
+                 n-word-bytes)))
+      (inst str value (@ thread-tn tls-index))))
+
+  (defun load-time-tls-index (symbol)
+    (let ((component (component-info *component-being-compiled*)))
+      (ash (or (position-if (lambda (x)
+                              (and (typep x '(cons (eql :tls-index)))
+                                   (eq (cdr x) symbol)))
+                            (ir2-component-constants component))
+               (vector-push-extend (cons :tls-index symbol)
+                                   (ir2-component-constants component)))
+           word-shift)))
+
+  (define-vop (bind)
+    (:args (value :scs (any-reg descriptor-reg zero) :to :save))
+    (:temporary (:sc non-descriptor-reg) tls-index)
+    (:temporary (:sc descriptor-reg) temp)
+    (:info symbol)
+    (:temporary (:scs (any-reg)) bsp)
+    (:generator 5
+      (load-binding-stack-pointer bsp)
+      (inst add bsp bsp (* binding-size n-word-bytes))
+      (store-binding-stack-pointer bsp)
+      (let* ((pkg (sb-xc:symbol-package symbol))
+             ;; These symbols should have a small enough index to be
+             ;; immediately encoded.
+             (known-symbol-p (or (and pkg (system-package-p pkg))
+                                 (eq pkg *cl-package*)))
              (tls-index-reg tls-index)
-             (tls-index (if known-symbol
-                            (make-fixup known-symbol :symbol-tls-index)
+             (tls-index (if known-symbol-p
+                            (make-fixup symbol :symbol-tls-index)
                             tls-index)))
         (assemble ()
-          (cond (known-symbol
+          (cond (known-symbol-p
                  (inst movz tls-index-reg tls-index))
                 (t
-                 (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
-                 (inst cbnz (32-bit-reg tls-index) TLS-INDEX-VALID)
-                 (move alloc-tls-symbol symbol)
-                 (load-inline-constant lr '(:fixup alloc-tls-index :assembly-routine))
-                 (inst blr lr)))
-          TLS-INDEX-VALID
-          (inst ldr alloc-tls-symbol (@ thread-tn tls-index))
-          (inst stp alloc-tls-symbol tls-index-reg
+                 ;; TODO: a fixup could replace this with an immediate
+                 (inst load-constant tls-index (load-time-tls-index symbol))))
+          (inst ldr temp (@ thread-tn tls-index))
+          (inst stp temp tls-index-reg
                 (@ bsp (* (- binding-value-slot binding-size)
                           n-word-bytes)))
           (inst str value (@ thread-tn tls-index))))))
