@@ -269,15 +269,13 @@
   (:temporary (:sc non-descriptor-reg) temp)
   (:conditional)
   (:info target not-p)
-  (:args-var args)
+  (:arg-refs integer-ref)
   (:policy :fast-safe)
   (:variant-vars comparison)
   (:variant :gt)
   (:generator 10
-    (unless (sc-is (tn-ref-tn args) descriptor-reg control-stack)
-      (setf args (tn-ref-across args)))
-    (let* ((integer-p (csubtypep (tn-ref-type args) (specifier-type 'integer)))
-           (other-pointer-p (fixnum-or-other-pointer-tn-ref-p args t))
+    (let* ((integer-p (csubtypep (tn-ref-type integer-ref) (specifier-type 'integer)))
+           (other-pointer-p (fixnum-or-other-pointer-tn-ref-p integer-ref t))
            negative-p
            (fixnum (if (sc-is fixnum immediate)
                        (let* ((value (fixnumize (tn-value fixnum)))
@@ -294,7 +292,7 @@
               (values not-target target)
               (values target not-target))
         (assemble ()
-          (when (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))
+          (when (types-equal-or-intersect (tn-ref-type integer-ref) (specifier-type 'fixnum))
             (cond ((or other-pointer-p
                        (not (and (eql fixnum 0)
                                  (eq comparison :ge))))
@@ -353,6 +351,7 @@
   (:args (fixnum :scs (immediate any-reg))
          (integer :scs (descriptor-reg)))
   (:arg-types tagged-num (:or integer bignum))
+  (:arg-refs nil integer-ref)
   (:variant :lt))
 
 (define-vop (<-fixnum-integer >-fixnum-integer)
@@ -573,7 +572,7 @@
 
 (define-vop (load-other-pointer-widetag)
   (:args (value :scs (any-reg descriptor-reg)))
-  (:args-var args)
+  (:arg-refs args)
   (:info not-other-pointer-label null-label)
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
@@ -591,11 +590,14 @@
   (:args (value :scs (unsigned-reg)))
   (:info target not-p type-codes)
   (:generator 1
-    (%test-headers value nil target not-p nil type-codes)))
+    (%test-headers value nil target not-p nil
+      (if (every #'integerp type-codes)
+          (canonicalize-widetags type-codes)
+          type-codes))))
 
 (define-vop (load-instance-layout)
   (:args (object :scs (any-reg descriptor-reg)))
-  (:args-var args)
+  (:arg-refs args)
   (:info not-instance)
   (:results (r :scs (descriptor-reg)))
   (:generator 1
@@ -668,7 +670,7 @@
   (:translate sb-c::structure-typep)
   (:args (object :scs (descriptor-reg)))
   (:arg-types * (:constant t))
-  (:args-var args)
+  (:arg-refs args)
   (:policy :fast-safe)
   (:conditional)
   (:info target not-p test-layout)
@@ -731,12 +733,11 @@
 
 (define-vop (keywordp type-predicate)
   (:translate keywordp)
-  (:args-var args-ref)
   (:generator 3
     #.(assert (= sb-impl::package-id-bits 16))
-    (unless (csubtypep (tn-ref-type args-ref) (specifier-type 'symbol))
+    (unless (csubtypep (tn-ref-type args) (specifier-type 'symbol))
       (test-type value temp (if not-p target not-target) t (symbol-widetag)
-                 :value-tn-ref args-ref))
+                 :value-tn-ref args))
     (inst ldrh temp (@ value (+ (ash symbol-name-slot word-shift)
                                (- other-pointer-lowtag)
                                6)))
