@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; Find the DFO for a component, deleting any unreached blocks and
 ;;; merging any other components we reach. We repeatedly iterate over
@@ -92,27 +92,15 @@
     (setf (block-flag block) t)
     (dolist (succ (block-succ block))
       (find-dfo-aux succ head component))
-    (when (component-nlx-info-generated-p component)
-      ;; FIXME: We also need (and do) this walk before physenv
-      ;; analysis, but at that time we are probably not very
-      ;; interested in the actual DF order.
-      ;;
-      ;; TODO: It is probable that one of successors have the same (or
-      ;; similar) set of NLXes; try to shorten the walk (but think
-      ;; about a loop, the only exit from which is non-local).
-      (map-block-nlxes (lambda (nlx-info)
-                         (let ((nle (nlx-info-target nlx-info)))
-                         (find-dfo-aux nle head component)))
-                       block))
     (remove-from-dfo block)
     (add-to-dfo block head))
   (values))
 
 ;;; This function is called on each block by FIND-INITIAL-DFO-AUX
-;;; before it walks the successors. It looks at the home CLAMBDA's
-;;; BIND block to see whether that block is in some other component:
+;;; before it walks the successors. It looks at the home lambda's
+;;; bind block to see whether that block is in some other component:
 ;;; -- If the block is in the initial component, then do
-;;;    DFO-SCAVENGE-DEPENDENCY-GRAPH on the home function to move it
+;;;    DFO-WALK-CALL-GRAPH on the home function to move it
 ;;;    into COMPONENT.
 ;;; -- If the block is in some other component, join COMPONENT into
 ;;;    it and return that component.
@@ -125,14 +113,14 @@
 ;;; the same component, even when they might not seem reachable from
 ;;; the environment entry. Consider the case of code that is only
 ;;; reachable from a non-local exit.
-(defun scavenge-home-dependency-graph (block component)
+(defun walk-home-call-graph (block component)
   (declare (type cblock block) (type component component))
   (let ((home-lambda (block-home-lambda block)))
     (if (eq (functional-kind home-lambda) :deleted)
         component
         (let ((home-component (lambda-component home-lambda)))
           (cond ((eq (component-kind home-component) :initial)
-                 (dfo-scavenge-dependency-graph home-lambda component))
+                 (dfo-walk-call-graph home-lambda component))
                 ((eq home-component component)
                  component)
                 (t
@@ -163,7 +151,7 @@
      ((block-flag block) component)
      (t
       (setf (block-flag block) t)
-      (let ((current (scavenge-home-dependency-graph block component)))
+      (let ((current (walk-home-call-graph block component)))
         (dolist (succ (block-succ block))
           (setq current (find-initial-dfo-aux succ current)))
         (remove-from-dfo block)
@@ -190,28 +178,17 @@
       (let* ((home (node-home-lambda ref))
              (home-kind (functional-kind home))
              (home-externally-visible-p
-              (or (eq home-kind :toplevel)
-                  (functional-has-external-references-p home)
-                  (let ((entry (functional-entry-fun home)))
-                    (and entry
-                         (functional-has-external-references-p entry))))))
+               (or (eq home-kind :toplevel)
+                   (functional-has-external-references-p home))))
         (unless (or (and home-externally-visible-p
                          (eq (functional-kind fun) :external))
                     (eq home-kind :deleted))
           (res home))))
     (res)))
 
-;;; If CLAMBDA is already in COMPONENT, just return that
-;;; component. Otherwise, move the code for CLAMBDA and all lambdas it
-;;; physically depends on (either because of calls or because of
-;;; closure relationships) into COMPONENT, or possibly into another
-;;; COMPONENT that we find to be related. Return whatever COMPONENT we
-;;; actually merged into.
-;;;
-;;; (Note: The analogous CMU CL code only scavenged call-based
-;;; dependencies, not closure dependencies. That seems to've been by
-;;; oversight, not by design, as per the bug reported by WHN on
-;;; cmucl-imp ca. 2001-11-29 and explained by DTC shortly after.)
+;;; Move the code for FUN and all functions called by it into
+;;; COMPONENT. If FUN is already in COMPONENT, just return that
+;;; component.
 ;;;
 ;;; If the function is in an initial component, then we move its head
 ;;; and tail to COMPONENT and add it to COMPONENT's lambdas. It is
@@ -219,16 +196,16 @@
 ;;; unreachable) because if the return is unreachable it (and its
 ;;; successor link) will be deleted in the post-deletion pass.
 ;;;
-;;; We then do a FIND-DFO-AUX starting at the head of CLAMBDA. If this
+;;; We then do a FIND-DFO-AUX starting at the head of FUN. If this
 ;;; flow-graph walk encounters another component (which can only
 ;;; happen due to a non-local exit), then we move code into that
 ;;; component instead. We then recurse on all functions called from
-;;; CLAMBDA, moving code into whichever component the preceding call
+;;; FUN, moving code into whichever component the preceding call
 ;;; returned.
 ;;;
-;;; If CLAMBDA is in the initial component, but the BLOCK-FLAG is set
-;;; in the bind block, then we just return COMPONENT, since we must
-;;; have already reached this function in the current walk (or the
+;;; If FUN is in the initial component, but the BLOCK-FLAG is set in
+;;; the bind block, then we just return COMPONENT, since we must have
+;;; already reached this function in the current walk (or the
 ;;; component would have been changed).
 ;;;
 ;;; If the function is an XEP, then we also walk all functions that
@@ -237,79 +214,48 @@
 ;;; ensures that conversion of a full call to a local call won't
 ;;; result in a need to join components, since the components will
 ;;; already be one.
-(defun dfo-scavenge-dependency-graph (clambda component)
-  (declare (type clambda clambda) (type component component))
-  (aver (not (eql (lambda-kind clambda) :deleted)))
-  (let* ((bind-block (node-block (lambda-bind clambda)))
-         (old-lambda-component (block-component bind-block))
-         (return (lambda-return clambda)))
+(defun dfo-walk-call-graph (fun component)
+  (declare (type clambda fun) (type component component))
+  (aver (not (eql (lambda-kind fun) :deleted)))
+  (let* ((bind-block (node-block (lambda-bind fun)))
+         (this (block-component bind-block))
+         (return (lambda-return fun)))
     (cond
-     ((eq old-lambda-component component)
+     ((eq this component)
       component)
-     ((not (eq (component-kind old-lambda-component) :initial))
-      (join-components old-lambda-component component)
-      old-lambda-component)
+     ((not (eq (component-kind this) :initial))
+      (join-components this component)
+      this)
      ((block-flag bind-block)
       component)
      (t
-      (push clambda (component-lambdas component))
-      (setf (component-lambdas old-lambda-component)
-            (delete clambda (component-lambdas old-lambda-component)))
+      (push fun (component-lambdas component))
+      (setf (component-lambdas this)
+            (delete fun (component-lambdas this)))
       (link-blocks (component-head component) bind-block)
-      (unlink-blocks (component-head old-lambda-component) bind-block)
+      (unlink-blocks (component-head this) bind-block)
       (when return
         (let ((return-block (node-block return)))
           (link-blocks return-block (component-tail component))
-          (unlink-blocks return-block (component-tail old-lambda-component))))
+          (unlink-blocks return-block (component-tail this))))
       (let ((res (find-initial-dfo-aux bind-block component)))
         (declare (type component res))
-        ;; Scavenge related lambdas.
-        (labels ((scavenge-lambda (clambda)
-                   (setf res
-                         (dfo-scavenge-dependency-graph (lambda-home clambda)
-                                                        res)))
-                 (scavenge-possibly-deleted-lambda (clambda)
-                   (unless (eql (lambda-kind clambda) :deleted)
-                     (scavenge-lambda clambda)))
-                 ;; Scavenge call relationship.
-                 (scavenge-call (called-lambda)
-                   (scavenge-lambda called-lambda))
-                 ;; Scavenge closure over a variable: if CLAMBDA
-                 ;; refers to a variable whose home lambda is not
-                 ;; CLAMBDA, then the home lambda should be in the
-                 ;; same component as CLAMBDA. (sbcl-0.6.13, and CMU
-                 ;; CL, didn't do this, leading to the occasional
-                 ;; failure when physenv analysis, which is local to
-                 ;; each component, would bogusly conclude that a
-                 ;; closed-over variable was unused and thus delete
-                 ;; it. See e.g. cmucl-imp 2001-11-29.)
-                 (scavenge-closure-var (var)
-                   (unless (null (lambda-var-refs var)) ; unless var deleted
-                     (let ((var-home-home (lambda-home (lambda-var-home var))))
-                       (scavenge-possibly-deleted-lambda var-home-home))))
-                 ;; Scavenge closure over an entry for nonlocal exit.
-                 ;; This is basically parallel to closure over a
-                 ;; variable above.
-                 (scavenge-entry (entry)
-                   (declare (type entry entry))
-                   (let ((entry-home (node-home-lambda entry)))
-                     (scavenge-possibly-deleted-lambda entry-home))))
-          (do-sset-elements (cc (lambda-calls-or-closes clambda))
-            (etypecase cc
-              (clambda (scavenge-call cc))
-              (lambda-var (scavenge-closure-var cc))
-              (entry (scavenge-entry cc))))
-          (when (eq (lambda-kind clambda) :external)
-            (mapc #'scavenge-call (find-reference-funs clambda))))
-        ;; Voila.
-        res)))))
+        (flet ((walk (fun)
+                 (unless (member (lambda-kind fun) '(:deleted :let :assignment))
+                   (setq res (dfo-walk-call-graph fun res)))))
+          (do-sset-elements (fun (lambda-calls fun))
+            (walk fun))
+          (when (eq (lambda-kind fun) :external)
+            (dolist (fun (find-reference-funs fun))
+              (walk fun)))
+          res))))))
 
-;;; Return true if CLAMBDA either is an XEP or has EXITS to some of
-;;; its ENTRIES.
-(defun has-xep-or-nlx (clambda)
-  (declare (type clambda clambda))
-  (or (eq (functional-kind clambda) :external)
-      (let ((entries (lambda-entries clambda)))
+;;; Return true if FUN either is an XEP or has EXITS to some of its
+;;; ENTRIES.
+(defun has-xep-or-nlx (fun)
+  (declare (type clambda fun))
+  (or (eq (functional-kind fun) :external)
+      (let ((entries (lambda-entries fun)))
         (and entries
              (find-if #'entry-exits entries)))))
 
@@ -317,13 +263,13 @@
 ;;; resulting components. Components with a :TOPLEVEL lambda, but no
 ;;; normal XEPs or potential non-local exits are marked as :TOPLEVEL.
 ;;; If there is a :TOPLEVEL lambda, and also a normal XEP, then we
-;;; treat the component as normal, but also return such components in
-;;; a list as the third value. Components with no entry of any sort
-;;; are deleted.
+;;; treat the component as normal. Components with no entry of any
+;;; sort are deleted.
 (defun separate-toplevelish-components (components)
   (declare (list components))
-  (collect ((non-top)
-            (top))
+  (collect ((real)
+            (top)
+            (real-top))
     (dolist (component components)
       (unless (eq (block-next (component-head component))
                   (component-tail component))
@@ -341,72 +287,65 @@
                  ;; references from pure :TOPLEVEL components. -- WHN
                  has-external-references
                  (setf (component-kind component) :complex-toplevel)
-                 (non-top component))
+                 (real component)
+                 (real-top component))
                 ((or (some #'has-xep-or-nlx funs)
                      (and has-top (rest funs)))
                  (setf (component-name component)
                        (possibly-base-stringize
                         (find-component-name component)))
-                 (non-top component)
+                 (real component)
                  (when has-top
-                   (setf (component-kind component) :complex-toplevel)))
+                   (setf (component-kind component) :complex-toplevel)
+                   (real-top component)))
                 (has-top
                  (setf (component-kind component) :toplevel)
                  (setf (component-name component) "top level form")
                  (top component))
                 (t
                  (delete-component component))))))
-    (values (non-top) (top))))
+    (values (real) (top) (real-top))))
 
-;;; Given a list of top level lambdas, return
-;;;   (VALUES NONTOP-COMPONENTS TOP-COMPONENTS):
-;;;   NONTOP-COMPONENTS = non-top-level-ish COMPONENTs;
-;;;   TOP-COMPONENTS = top-level-ish COMPONENTs;
+;;; Given a list of top-level lambdas, return three lists of components
+;;; representing the actual component division:
+;;;  1] the non-top-level components,
+;;;  2] and the second is the top-level components, and
+;;;  3] Components in [1] that also have a top-level lambda.
 ;;;
 ;;; We assign the DFO for each component, and delete any unreachable
 ;;; blocks. We assume that the FLAGS have already been cleared.
-(defun find-initial-dfo (toplevel-lambdas)
-  (declare (list toplevel-lambdas))
+;;;
+;;; We iterate over the lambdas in each initial component, trying
+;;; to put each function in its own component, but joining it to
+;;; an existing component if we find that there are references
+;;; between them. Any code that is left in an initial component
+;;; must be unreachable, so we can delete it. Stray links to the
+;;; initial component tail (due to NIL function terminated blocks)
+;;; are moved to the appropriate new component tail.
+(defun find-initial-dfo (top-level-lambdas)
+  (declare (list top-level-lambdas))
   (collect ((components))
-    ;; We iterate over the lambdas in each initial component, trying
-    ;; to put each function in its own component, but joining it to
-    ;; an existing component if we find that there are references
-    ;; between them. Any code that is left in an initial component
-    ;; must be unreachable, so we can delete it. Stray links to the
-    ;; initial component tail (due to NIL function terminated blocks)
-    ;; are moved to the appropriate new component tail.
-    (dolist (toplevel-lambda toplevel-lambdas)
-      (let* ((old-component (lambda-component toplevel-lambda))
-             (old-component-lambdas (component-lambdas old-component))
-             (new-component nil))
-        (aver (member toplevel-lambda old-component-lambdas))
-        (dolist (component-lambda old-component-lambdas)
-          (aver (member (functional-kind component-lambda)
-                        '(:optional :external :toplevel nil :escape
-                                    :cleanup)))
-          (unless new-component
-            (setf new-component (make-empty-component))
-            (setf (component-name new-component)
-                  ;; This isn't necessarily an ideal name for the
-                  ;; component, since it might end up with multiple
-                  ;; lambdas in it, not just this one, but it does
-                  ;; seem a better name than just "<unknown>".
-                  (leaf-debug-name component-lambda)))
-          (let ((res (dfo-scavenge-dependency-graph component-lambda
-                                                    new-component)))
-            (when (eq res new-component)
-              (aver (not (position new-component (components))))
-              (components new-component)
-              (setq new-component nil))))
-        (when (eq (component-kind old-component) :initial)
-          (aver (null (component-lambdas old-component)))
-          (let ((tail (component-tail old-component)))
+    (let ((new (make-empty-component)))
+      (dolist (initial-component (mapcar #'lambda-component top-level-lambdas))
+        (unless (eq (component-kind initial-component) :deleted)
+          (dolist (component-lambda (component-lambdas initial-component))
+            (aver (member (functional-kind component-lambda)
+                          '(:optional :external :toplevel nil :escape
+                            :cleanup)))
+            (let ((res (dfo-walk-call-graph component-lambda new)))
+              (when (eq res new)
+                (aver (not (member new (components))))
+                (components new)
+                (setq new (make-empty-component)))))
+          (aver (eq (component-kind initial-component) :initial))
+          (aver (null (component-lambdas initial-component)))
+          (let ((tail (component-tail initial-component)))
             (dolist (pred (block-pred tail))
               (let ((pred-component (block-component pred)))
-                (unless (eq pred-component old-component)
+                (unless (eq pred-component initial-component)
                   (unlink-blocks pred tail)
                   (link-blocks pred (component-tail pred-component))))))
-          (delete-component old-component))))
+          (delete-component initial-component))))
 
     ;; When we are done, we assign DFNs.
     (dolist (component (components))
@@ -415,6 +354,9 @@
         (do-blocks-backwards (block component :both)
           (setf (block-number block) (incf num)))))
 
+    (dolist (component (components))
+      (delete-if-no-entries component))
+
     ;; Pull out top-level-ish code.
     (separate-toplevelish-components (components))))
 
@@ -426,7 +368,7 @@
   (setf (functional-kind lambda) :deleted)
   (dolist (let (lambda-lets lambda))
     (setf (lambda-home let) result-lambda)
-    (setf (lambda-physenv let) (lambda-physenv result-lambda))
+    (setf (lambda-environment let) (lambda-environment result-lambda))
     (push let (lambda-lets result-lambda)))
   (setf (lambda-entries result-lambda)
         (nconc (lambda-entries result-lambda)
@@ -499,14 +441,7 @@
          (result-return (lambda-return result-lambda)))
     (cond
       ((null (rest lambdas)))
-      ((and result-return
-            ;; KLUDGE: why is the node deleted but clambda still has a return?
-            ;; see a test-case for this in tests/merge-lambdas.lisp
-            ;; But now it can only be exercised with block
-            ;; compilation, which doesn't seem to work anyway.
-            (not (node-to-be-deleted-p
-                  (ctran-use (node-prev
-                              (lvar-uses (return-result result-return)))))))
+      (result-return
        ;; Make sure the result's return node starts a block so that we
        ;; can splice code in before it.
        (let ((prev (node-prev
