@@ -123,8 +123,7 @@
     (do-nodes-backwards (node lvar block :restart-p t)
       (let ((dest (when lvar (lvar-dest lvar))))
         (cond ((and (cast-p dest)
-                    (not (cast-type-check dest))
-                    (almost-immediately-used-p lvar node))
+                    (not (cast-type-check dest)))
                (let ((dtype (node-derived-type node))
                      (atype (node-derived-type dest)))
                  (when (values-types-equal-or-intersect
@@ -144,13 +143,18 @@
                (delete-filter node lvar (cast-value node))))))))
 
 (defglobal *two-arg-functions*
-    `((* two-arg-*)
+    `((* two-arg-*
+         ,@(sb-c::unless-vop-existsp (:translate sb-c::fixnum*)
+             `((,(specifier-type 'fixnum) ,(specifier-type 'fixnum))
+               multiply-fixnums)))
       (+ two-arg-+)
       (- two-arg--)
       (/ two-arg-/ (,(specifier-type 'integer) ,(specifier-type 'integer)) sb-kernel::integer-/-integer)
       (< two-arg-<)
       (= two-arg-=)
       (> two-arg->)
+      (<= two-arg-<=)
+      (>= two-arg->=)
       (char-equal two-arg-char-equal)
       (char-greaterp two-arg-char-greaterp)
       (char-lessp two-arg-char-lessp)
@@ -192,7 +196,29 @@
 (def-two-arg-funs (character character)
   char= char/= char< char> char<= char>=)
 (def-two-arg-funs (number number)
-  >= <= /=)
+  /=)
+
+;;; A list of function which always call their functional argument correctly,
+;;; meaning that no arg-count-error can occur in the callee
+;;; and therefore the callee can skip the check.
+(dolist (fun '(sb-thread::call-with-mutex
+               sb-thread::call-with-recursive-lock
+               sb-thread::call-with-system-mutex
+               sb-thread::call-with-system-mutex/allow-with-interrupts
+               sb-thread::call-with-system-mutex/without-gcing
+               sb-thread::call-with-recursive-system-lock
+               ;; wish we had a little more commonality to these names
+               sb-impl::%with-standard-io-syntax
+               sb-impl::%with-rebound-io-syntax
+               sb-debug::funcall-with-debug-io-syntax
+               sb-impl::call-with-sane-io-syntax
+               ;; there are others, maybe %with-compilation-unit
+               sb-impl::%print-unreadable-object
+               ))
+  (let ((info (info :function :info fun)))
+    (when info
+      (setf (ir1-attributep (fun-info-attributes info) callee-omit-arg-count-check)
+            t))))
 
 ;;; Convert function designators to functions in calls to known functions
 ;;; Also convert to TWO-ARG- variants
@@ -201,44 +227,70 @@
     (do-nodes (node nil block)
       (when (and (combination-p node)
                  (eq (combination-kind node) :known)
+                 ;; REDUCE can call with zero arguments.
                  (neq (lvar-fun-name (combination-fun node) t) 'reduce))
         (map-callable-arguments
-           (lambda (lvar args results &key no-function-conversion &allow-other-keys)
-             (declare (ignore results))
-             (unless no-function-conversion
-               (let ((ref (lvar-uses lvar))
-                     (arg-count (length args)))
-                 (labels ((translate-two-args (name)
-                            (and (eql arg-count 2)
-                                 (not (fun-lexically-notinline-p name (node-lexenv node)))
-                                 (cadr (assoc name *two-arg-functions*))))
-                          (translate (ref)
-                            (let* ((leaf (ref-leaf ref))
-                                   (fun-name (and (constant-p leaf)
-                                                  (constant-value leaf)))
-                                   (replacement
-                                     (cond ((and fun-name (symbolp fun-name))
-                                            (translate-two-args fun-name))
-                                           ((and (global-var-p leaf)
-                                                 (eq (global-var-kind leaf) :global-function))
-                                            (translate-two-args (global-var-%source-name leaf)))))
-                                   (*compiler-error-context* node))
-                              (and replacement
-                                   (find-free-fun replacement "ir1-finalize")))))
-                   (cond ((ref-p ref)
-                          (let ((replacement (translate ref)))
-                            (when replacement
-                              (change-ref-leaf ref replacement))))
-                         ((cast-p ref)
-                          (let* ((cast ref)
-                                 (ref (lvar-uses (cast-value cast))))
-                            (when (ref-p ref)
-                              (let ((replacement (translate ref)))
-                                (when replacement
-                                  (change-ref-leaf ref replacement :recklessly t)
-                                  (setf (node-derived-type cast)
-                                        (lvar-derived-type (cast-value cast)))))))))))))
-           node)))))
+         (lambda (lvar args results &key no-function-conversion &allow-other-keys)
+           (declare (ignore results))
+           ;; Process annotations while the original values are still there.
+           (process-annotations lvar)
+           (unless no-function-conversion
+             (let ((ref (lvar-uses lvar))
+                   (arg-count (length args)))
+               (labels ((translate-two-args (name)
+                          (and (eql arg-count 2)
+                               (not (fun-lexically-notinline-p name (node-lexenv node)))
+                               (cadr (assoc name *two-arg-functions*))))
+                        (translate (ref)
+                          (let* ((leaf (ref-leaf ref))
+                                 (fun-name (and (constant-p leaf)
+                                                (constant-value leaf)))
+                                 (replacement
+                                   (cond ((and fun-name (symbolp fun-name))
+                                          (or (translate-two-args fun-name)
+                                              (and (not (memq (info :function :kind fun-name)
+                                                              '(:macro :special-form)))
+                                                   fun-name)))
+                                         ((and (global-var-p leaf)
+                                               (eq (global-var-kind leaf) :global-function))
+                                          (translate-two-args (global-var-%source-name leaf)))))
+                                 (*compiler-error-context* node))
+                            (and replacement
+                                 (find-global-fun replacement t)))))
+                 (cond ((ref-p ref)
+                        (let ((replacement (translate ref)))
+                          (when replacement
+                            (change-ref-leaf ref replacement))))
+                       ((cast-p ref)
+                        (let* ((cast ref)
+                               (ref (lvar-uses (cast-value cast))))
+                          (when (ref-p ref)
+                            (let ((replacement (translate ref)))
+                              (when replacement
+                                (change-ref-leaf ref replacement :recklessly t)
+                                (setf (node-derived-type cast)
+                                      (lvar-derived-type (cast-value cast)))))))))))))
+         node)
+        ;; One more thing: builtin higher-order functions utilized by builtin macros
+        ;; can impart a policy change to the callee, but it can't (easily) be done
+        ;; strictly lexically, because the policy would leak downward.
+        ;; e.g. (WITH-SOME-STUFF () ..) ->
+        ;; -> (DX-FLET ((THUNK () <user-code>)) (CALL-WITH-STUFF #'THUNK))
+        ;; should not inject (OPTIMIZE (VERIFY-ARG-COUNT 0)) at the top of <user-code>
+        ;; because every lambda therein would omit the arg-count check.
+        (when (ir1-attributep (fun-info-attributes (combination-fun-info node))
+                              callee-omit-arg-count-check)
+          (let* ((args (combination-args node))
+                 (arg (if (eq (lvar-fun-name (combination-fun node) t)
+                              'sb-impl::%print-unreadable-object)
+                          (fourth args)
+                          (first args)))
+                 (ref (and arg (lvar-uses arg))))
+            (when (and (ref-p ref) (lambda-p (ref-leaf ref)))
+              ;; It seems impolite (un-debuggable too) to alter the lexenv-policy
+              ;; of functional-lexenv, so annotate this differently.
+              (setf (getf (functional-plist (ref-leaf ref)) 'verify-arg-count)
+                    nil))))))))
 
 (defun rewrite-full-call (combination)
   (let ((combination-name (lvar-fun-name (combination-fun combination) t))
@@ -257,6 +309,8 @@
                            for type in types
                            always (csubtypep (lvar-type arg) type)))
             (setf two-arg typed-two-arg))
+          (setf (combination-fun-info combination)
+                (fun-info-or-lose two-arg))
           (change-ref-leaf
            ref
            (find-free-fun two-arg "rewrite-full-call")))))))
