@@ -154,7 +154,22 @@
   (declare (type ref ref) (type combination call) (type clambda fun))
   (propagate-to-args call fun)
   (setf (basic-combination-kind call) :local)
-  (sset-adjoin fun (lambda-calls-or-closes (node-home-lambda call)))
+  (let ((fun-environment (lambda-environment fun))
+        (call-home (node-home-lambda call)))
+    ;; FUN might close over top-level variables. Make sure we
+    ;; propagate those in the local call environment.
+    (when fun-environment
+      (dolist (thing (environment-closure fun-environment))
+        (let ((thing-env
+                (lambda-environment
+                 (etypecase thing
+                   (lambda-var (lambda-var-home thing))
+                   (nlx-info (node-home-lambda (exit-entry (nlx-info-exit thing))))
+                   (clambda
+                    (aver (xep-p thing))
+                    (node-home-lambda (xep-enclose thing)))))))
+          (close-over thing (lambda-environment call-home) thing-env))))
+    (sset-adjoin fun (lambda-calls call-home)))
   (recognize-potentially-dynamic-extent-lvars call fun)
   (merge-tail-sets call fun)
   (change-ref-leaf ref fun)
@@ -603,7 +618,7 @@
                      (call-all-args-fixed-p call)))
         (aver (= (optional-dispatch-min-args fun) 0))
         (setf (basic-combination-kind call) :local)
-        (sset-adjoin ep (lambda-calls-or-closes (node-home-lambda call)))
+        (sset-adjoin ep (lambda-calls (node-home-lambda call)))
         (merge-tail-sets call ep)
         (change-ref-leaf ref ep)
         (if (singleton-p args)
@@ -1034,13 +1049,12 @@
     (setf (lambda-lets clambda) nil)
 
     ;; HOME no longer calls CLAMBDA, and owns all of CLAMBDA's old
-    ;; DFO dependencies.
-    (sset-union (lambda-calls-or-closes home)
-                (lambda-calls-or-closes clambda))
-    (sset-delete clambda (lambda-calls-or-closes home))
+    ;; calls.
+    (sset-union (lambda-calls home) (lambda-calls clambda))
+    (sset-delete clambda (lambda-calls home))
     ;; CLAMBDA no longer has an independent existence as an entity
-    ;; which calls things or has DFO dependencies.
-    (setf (lambda-calls-or-closes clambda) nil)
+    ;; which calls things.
+    (setf (lambda-calls clambda) nil)
     ;; Make sure the exits that are no longer non-local are deleted
     (loop for entry in (lambda-entries home)
           do (loop for exit in (entry-exits entry)
@@ -1100,40 +1114,39 @@
 ;;; all calls were TR.)
 (defun unconvert-tail-calls (fun call next-block)
   (let (maybe-terminate)
-    (do-sset-elements (called (lambda-calls-or-closes fun))
-      (when (lambda-p called)
-        (dolist (ref (leaf-refs called))
-          (let ((this-call (node-dest ref)))
-            (when (and this-call
-                       (node-tail-p this-call)
-                       (not (node-to-be-deleted-p this-call))
-                       (eq (node-home-lambda this-call) fun))
-              (setf (node-tail-p this-call) nil)
-              (ecase (functional-kind called)
-                ((nil :cleanup :optional)
-                 (let ((block (node-block this-call))
-                       (lvar (node-lvar call)))
-                   (unlink-blocks block (first (block-succ block)))
-                   (link-blocks block next-block)
-                   (if (eq (node-derived-type this-call) *empty-type*)
-                       ;; Delay terminating the block, because there may be more calls
-                       ;; to be processed here and this may prematurely delete NEXT-BLOCK
-                       ;; before we attach more preceding blocks to it.
-                       ;; Although probably if one call to a function
-                       ;; is derived to be NIL all other calls would
-                       ;; be NIL too, but that may not be available at the same time.
-                       ;; (Or something is smart in the future to
-                       ;; derive different results from different
-                       ;; calls.)
-                       (push this-call maybe-terminate)
-                       (add-lvar-use this-call lvar))))
-                (:deleted)
-                ;; The called function might be an assignment in the
-                ;; case where we are currently converting that function.
-                ;; In steady-state, assignments never appear as a called
-                ;; function.
-                (:assignment
-                 (aver (eq called fun)))))))))
+    (do-sset-elements (called (lambda-calls fun))
+      (dolist (ref (leaf-refs called))
+        (let ((this-call (node-dest ref)))
+          (when (and this-call
+                     (node-tail-p this-call)
+                     (not (node-to-be-deleted-p this-call))
+                     (eq (node-home-lambda this-call) fun))
+            (setf (node-tail-p this-call) nil)
+            (ecase (functional-kind called)
+              ((nil :cleanup :optional)
+               (let ((block (node-block this-call))
+                     (lvar (node-lvar call)))
+                 (unlink-blocks block (first (block-succ block)))
+                 (link-blocks block next-block)
+                 (if (eq (node-derived-type this-call) *empty-type*)
+                     ;; Delay terminating the block, because there may be more calls
+                     ;; to be processed here and this may prematurely delete NEXT-BLOCK
+                     ;; before we attach more preceding blocks to it.
+                     ;; Although probably if one call to a function
+                     ;; is derived to be NIL all other calls would
+                     ;; be NIL too, but that may not be available at the same time.
+                     ;; (Or something is smart in the future to
+                     ;; derive different results from different
+                     ;; calls.)
+                     (push this-call maybe-terminate)
+                     (add-lvar-use this-call lvar))))
+              (:deleted)
+              ;; The called function might be an assignment in the
+              ;; case where we are currently converting that function.
+              ;; In steady-state, assignments never appear as a called
+              ;; function.
+              (:assignment
+               (aver (eq called fun))))))))
     maybe-terminate))
 
 ;;; Deal with returning from a LET or assignment that we are
@@ -1446,6 +1459,7 @@
              (not (block-delete-p (lambda-block fun))))
     (let ((outside-calls nil)
           (outside-calls-ctran nil)
+          (outside-calls-lvar nil)
           (outside-calls-env nil)
           (outside-calls-cleanup nil))
       (when (and (dolist (ref (leaf-refs fun) t)
@@ -1461,6 +1475,9 @@
                            (let ((dest-ctran
                                    (or (node-next dest)
                                        (block-start (first (block-succ (node-block dest))))))
+                                 (dest-lvar
+                                   (and (valued-node-p dest)
+                                        (node-lvar dest)))
                                  (dest-env
                                    (node-home-lambda dest))
                                  (dest-cleanup
@@ -1481,13 +1498,15 @@
                                     ;; cleanups with respect to the
                                     ;; messiest common ancestor.
                                     (unless (and (or (eq (node-derived-type dest) *empty-type*)
-                                                     (and (eq outside-calls-ctran dest-ctran)))
+                                                     (and (eq outside-calls-ctran dest-ctran)
+                                                          (eq outside-calls-lvar dest-lvar)))
                                                  (eq outside-calls-env dest-env)
                                                  (eq outside-calls-cleanup dest-cleanup))
                                       (return nil)))
                                    (t
                                     (setq outside-calls-env dest-env)
                                     (setq outside-calls-ctran dest-ctran)
+                                    (setq outside-calls-lvar dest-lvar)
                                     (setq outside-calls-cleanup dest-cleanup)))
                              (push dest outside-calls))))))
                  (ok-initial-convert-p fun))
